﻿using System;
using System.Threading;

using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

using GHIElectronics.NETMF.FEZ;
using GHIElectronics.NETMF.Net;
using GHIElectronics.NETMF.Net.NetworkInformation;
using System.IO;
using System.Text;

namespace BeerTemperature
{
    public class Program
    {
        private static FEZ_Components.Thermometer _thermometer;
       
        private static string _data = string.Empty;
        private static Stream _s = null;
        private static HttpWebResponse _response = null;
        private static HttpWebRequest _request = null;

        // Klíč Cosm
        private static string API_KEY = "";
        // ID kanálu
        private static string FEED_ID = "";

        // displej FEZ Touch
        private static FEZ_Components.FEZTouch _fezTouch;
        private static FEZ_Components.FEZTouch.Font _fontArial16;
        private static FEZ_Components.FEZTouch.Color _backgroundColor;

        public static void Main()
        {

            _thermometer = new FEZ_Components.Thermometer(FEZ_Pin.AnalogIn.An2);
                       
            // inicializace FEZ Connect Shield
            WIZnet_W5100.Enable(SPI.SPI_module.SPI1, (Cpu.Pin)FEZ_Pin.Digital.Di10, (Cpu.Pin)FEZ_Pin.Digital.Di7, true);
            // Povolení DHCP, je třeba mít MAC adresu. Nejjednudušší varinata je nákup staré síťové karty v bazaru, nebo použití MAC adresy z již vyřazené karty
            Dhcp.EnableDhcp(new byte[] { 0x00, 0x3F, 0x3F, 0xED, 0xAA, 0x2B }, "FEZ");
            Dhcp.RenewDhcpLease();

            InitGraphics();

            PrintCaption();


            Debug.Print(new IPAddress(NetworkInterface.IPAddress).ToString());
            Debug.Print(new IPAddress(NetworkInterface.DnsServer).ToString());

            PrintStatus(Status.Ready);

            Thread.Sleep(Timeout.Infinite);
        }

        private static void PrintCaption()
        {
            PrintText("Beer temperature", FEZ_Components.FEZTouch.Color.White, 20);
        }

        private static void PrintTemperature(float temperature)
        {
            PrintText(temperature.ToString(), FEZ_Components.FEZTouch.Color.Red, 80);
        }

        private static void PrintText(string text, FEZ_Components.FEZTouch.Color color, int topOffset)
        {
            string[] texts = text.Split(new char[] { '\n' });

            int posX = 0;
            int iterator = 0;

            foreach (var item in texts)
            {                
                posX = (_fezTouch.ScreenWidth - _fontArial16.GetTextWidth(item)) / 2;
                _fezTouch.FillRectangle(0, topOffset + ((_fontArial16.Height + 3) * iterator), _fezTouch.ScreenWidth, _fontArial16.Height, _backgroundColor);
                _fezTouch.DrawString(posX, topOffset + ((_fontArial16.Height + 3) * iterator++), item, color, _backgroundColor, _fontArial16);    
            }            
        }

        enum Status { Wait, OK, Error, Ready, Saving };

        private static void PrintStatus(Status status)
        {
            string statusText = "Error\n ";
            FEZ_Components.FEZTouch.Color statusColor = FEZ_Components.FEZTouch.Color.Red;

            switch (status)
            {
                case Status.OK:
                    statusText = "Saved\n ";
                    statusColor = FEZ_Components.FEZTouch.Color.Green;
                    break;

                case Status.Wait:
                    statusText = "Wait...\n ";
                    statusColor = FEZ_Components.FEZTouch.Color.Yellow;
                    break;

                case Status.Saving:
                    statusText = "Saving to Cosm\n ";
                    statusColor = FEZ_Components.FEZTouch.Color.Yellow;
                    break;

                case Status.Ready:
                    statusText = "Put sensor to beer\ntouch display...";
                    statusColor = FEZ_Components.FEZTouch.Color.White;
                    break;
            }

            PrintText(statusText, statusColor, 140);
        }

        
        public static void InitGraphics()
        {
            // create lcd configuration for FEZ Panda II
            FEZ_Components.FEZTouch.LCDConfiguration lcdConfig = new FEZ_Components.FEZTouch.LCDConfiguration(
                FEZ_Pin.Digital.Di28,
                FEZ_Pin.Digital.Di20,
                FEZ_Pin.Digital.Di22,
                FEZ_Pin.Digital.Di23,
                new FEZ_Pin.Digital[8] { FEZ_Pin.Digital.Di51, FEZ_Pin.Digital.Di50, FEZ_Pin.Digital.Di49, FEZ_Pin.Digital.Di48, FEZ_Pin.Digital.Di47, FEZ_Pin.Digital.Di46, FEZ_Pin.Digital.Di45, FEZ_Pin.Digital.Di44 },
                FEZ_Pin.Digital.Di24,
                FEZ_Pin.Digital.Di26,
                FEZ_Components.FEZTouch.Orientation.LandscapeInverse
                );

            // create touch configuration for FEZ Panda II
            FEZ_Components.FEZTouch.TouchConfiguration touchConfig = new FEZ_Components.FEZTouch.TouchConfiguration(SPI.SPI_module.SPI2, FEZ_Pin.Digital.Di25, FEZ_Pin.Digital.Di34);

            // create fez touch
            _fezTouch = new FEZ_Components.FEZTouch(lcdConfig, touchConfig);            

            // font
            _fontArial16 = new FEZTouch.Fonts.FontCourierNew18();

            _backgroundColor = FEZ_Components.FEZTouch.Color.Black;
            
            // clear the screen
            _fezTouch.ClearScreen();

            _fezTouch.TouchDownEvent += new FEZ_Components.FEZTouch.TouchEventHandler(_fezTouch_TouchDownEvent);
        }

        static void _fezTouch_TouchDownEvent(int x, int y)
        {
            // temperování senzoru
            PrintStatus(Status.Wait);
            Thread.Sleep(5000);
            

            // Požadavek na odeslání dat na Pachube
            using (_request = (HttpWebRequest)HttpWebRequest.Create("http://api.pachube.com/v2/feeds/" + FEED_ID + ".csv"))
            {
                _request.Method = "PUT";

                // Nejjednodušší formát dat
                _request.ContentType = "text/csv";

                // Autorizační klíč
                _request.Headers.Add("X-PachubeApiKey", API_KEY);

                float temperature = _thermometer.GetTemperatureCelsius();

                // Získání teploty
                _data = "temperature," + temperature;
                Debug.Print(_data);

                PrintTemperature(temperature);


                byte[] buffer = Encoding.UTF8.GetBytes(_data);

                _request.ContentLength = buffer.Length;

                _request.Timeout = 5000;

                PrintStatus(Status.Saving);

                try
                {
                    _s = _request.GetRequestStream();
                    _s.Write(buffer, 0, buffer.Length);

                    using (_response = (HttpWebResponse)_request.GetResponse())
                    {
                        Debug.Print("Status code: " + _response.StatusCode);
                        if (_response.StatusCode == HttpStatusCode.OK)
                        {
                            PrintStatus(Status.OK);
                        }
                        else
                        {
                            PrintStatus(Status.Error);
                        }
                    }
                }
                catch
                {
                    PrintStatus(Status.Error);
                }
                finally
                {
                    Thread.Sleep(3000);
                    PrintStatus(Status.Ready);
                }
            }
        }

    }
}
